
import java.io.*;
import java.util.*;
import java.awt.*;
import josx.rcxcomm.*;




class AlgoRechercheChemin
{
	/*INVARIANTS DE TYPES*/
	/*@
	  @ private invariant (\forall int i; i >=0 && i < obstacles.size(); obstacles.get(i) instanceof Obstacle);
	  @*/
	private Vector obstacles = null;

	/*@
	  @ private invariant (\forall int i; i >=0 && i < refs.size(); refs.get(i) instanceof ObstacleReference);
	  @*/
	private Vector refs = null;

	/*@
	  @ private invariant (\forall int i; i >=0 && i < aretes.size(); aretes.get(i) instanceof Arete);
	  @*/
	private Vector aretes = null;

	/*@
	  @ private invariant (\forall int i; i >=0 && i < chemin.size(); chemin.get(i) instanceof Arete);
	  @*/
	private Vector chemin = null;

	/*@
	  @ private invariant (\forall int i; i >=0 && i < milieux.size(); milieux.get(i) instanceof Point);
	  @*/
	private Vector milieux = null;


	RCXPort port;

	InputStream is;
	OutputStream os;
	DataInputStream dis;
	DataOutputStream dos;


	AlgoRechercheChemin()
	{
	/* initialisations */

	obstacles = new Vector();
	refs = new Vector();
	aretes = new Vector();
	chemin = new Vector();
	milieux = new Vector();
	try
		{
		port = new RCXPort();
		is = port.getInputStream();
		os = port.getOutputStream();
		dis = new DataInputStream(is);
		dos = new DataOutputStream(os);
		}
	catch (Exception e)
		{}

	/* tapes de droulement de l'algorithme */

	creerEnvironnement();
	genererReferences();
	genererAretes();
	genererZones();
	construireGraphe();

	Arete courante = ((Cadre) obstacles.get(0)).getArete(0);
	getChemin(courante);
	getMilieux();
	transmetRCX();
	recupereRCX();
	}


	/* invariant : aucune canette ne se chevauche ou ne sort du cadre
	 * postcondition : il existe au moins un cadre
	 */
	/*@
	  @ private invariant (\forall int i;i>=0 && i<obstacles.size();
	  @				  	   (\forall int j; j >= 0 && j < obstacles.size() && i != j;
	  @							!((Obstacle) obstacles.get(i)).chevauche((Obstacle) obstacles.get(j))));
	  @
	  @ private normal_behavior
	  @ ensures (\exists int x; x >= 0 && x < obstacles.size();
	  @			    (obstacles.get(x) instanceof Cadre));
	  @*/
	private void creerEnvironnement()
	{
	ajouterCadre(120, 50);
	ajouterCanette(new Point(30,15), 3);
	ajouterCanette(new Point(40,36), 3);
	ajouterCanette(new Point(90,15), 3);
	}


	/* prcondition : il n'existe pas de cadre
	 * postcondition : il existe un cadre unique
	 */
	/*@
	  @ requires (\forall int i; 0 <= i && i < obstacles.size();
	  @		   !(obstacles.get(i) instanceof Cadre));
	  @
	  @ ensures (\sum int x; x >= 0 && x < obstacles.size() && (obstacles.get(x) instanceof Cadre) ; 1)==1;
	  @*/
	private void ajouterCadre(int longueur, int hauteur)
	{
	obstacles.add(new Cadre(new Point(0,0), hauteur, longueur));
 	}


	/* prcondition : il existe un cadre unique
	 */
	/*@
	  @ requires (\sum int x; x >= 0 && x < obstacles.size() && (obstacles.get(x) instanceof Cadre) ; 1)==1;
	  @*/
	private void ajouterCanette(Point origine, int rayon)
	{
	obstacles.add(new Canette(origine, rayon));
 	}


	private void genererReferences()
	{
	Object array[];

	for (int i = 0; i < obstacles.size(); i++)
		{
		Obstacle o;

		o = (Obstacle) obstacles.get(i);
		refs.addAll(o.getReferences());
		}

	array = refs.toArray();
	Arrays.sort(array);
	int size = refs.size();
	refs.clear();
	for (int i = 0; i < size; i++)
		{
		refs.add(array[i]);
		}
	}


	private void genererAretes()
	{
	for (int i = 1; i < (refs.size()-1); i++)
		{
		Obstacle o = ((ObstacleReference) refs.get(i)).getObstacle();
		Point p = ((ObstacleReference) refs.get(i)).getPoint();
		Point toto;
		ObstacleReference current = null;
		int j;
		boolean trouve;
		Arete arete;


		ObstacleReference refBas = (ObstacleReference) refs.firstElement();
		j = i-1;
		trouve = false;
		while ((j >= 0) && (!trouve))
			{
			current = (ObstacleReference) refs.get(j);
			if ((current.getObstacle().intersection(p)) && (p.getY() >  current.getPoint().getY()) && (refBas.getPoint().getY() <=  current.getPoint().getY()))
				{
				toto = new Point((int) ((ObstacleReference) refs.get(i)).getPoint().getX(), 1);
				trouve = true;

				arete = new Arete((ObstacleReference) refs.get(i), new ObstacleReference(toto, current.getObstacle()));

				((ObstacleReference) refs.get(i)).getObstacle().addArete(arete);

				if (!(((ObstacleReference) refs.get(i)).getObstacle() instanceof Cadre))
					{
					current.getObstacle().addArete(arete);
					}
				}
			j--;
			}

		ObstacleReference refHaut = (ObstacleReference) refs.lastElement();
		j = i+1;
		trouve = false;
		while ((j < refs.size()) && (!trouve))
			{
			current = (ObstacleReference) refs.get(j);
			if ((current.getObstacle().intersection(p)) && (p.getY() <  current.getPoint().getY()) && (refHaut.getPoint().getY() >=  current.getPoint().getY()))
				{
				toto = new Point((int) ((ObstacleReference) refs.get(i)).getPoint().getX(), -1);
				trouve = true;

				arete = new Arete((ObstacleReference) refs.get(i), new ObstacleReference(toto, current.getObstacle()));

				/*  prvention des doublons */
				if ((current.getObstacle() instanceof Cadre) || (current.getObstacle().notContains(((ObstacleReference) refs.get(i)).getPoint().getX())))
					{
					((ObstacleReference) refs.get(i)).getObstacle().addArete(arete);
					}

				if (!(((ObstacleReference) refs.get(i)).getObstacle() instanceof Cadre) && (((ObstacleReference) refs.get(i)).getPoint().getX() != current.getPoint().getX()))
					{
					current.getObstacle().addArete(arete);
					}
				}
			j++;
			}
		}
	}


	private void afficherAretes()
	{
	for (int z = 0; z < obstacles.size(); z++)
		{
		System.out.println("Obstacle num " + z + " :");
		for (int y = 0; y < ((Obstacle) obstacles.get(z)).nbArete(); y++)
			{
			System.out.println(((Obstacle) obstacles.get(z)).getArete(y).toString());
			}
		System.out.println("");
		}
	}


	private void genererZones()
	{
	for (int y = 0; y < obstacles.size(); y++)
		{
		((Obstacle) obstacles.get(y)).triAretes();
		}
	}


	private void rechSuccesseurs(Arete depart, int sensInitial)
	{
	Arete curArr = depart;
	int sens = sensInitial;

	do
		{
		Obstacle destObs;

		/* objet de destination */
		if (sens > 0)
			destObs = curArr.getObsDest();
		else
			destObs = curArr.getObsSrc();

		/* retrouve l'arete courante sur objet de destination */
		int ia = 0;
		while (destObs.getArete(ia) != curArr)
			{
			ia++;
			}

		/* on trouve la prochaine arete */
		ia--;
		if (ia == -1)
			ia = destObs.nbArete()-1;

		/* calcule la prochaine arete et son sens de parcours */
		curArr = destObs.getArete(ia);
		if (curArr.getObsSrc() == destObs)
			sens = 1;
		else
			sens = -1;

		/* l'ajoute comme successeur de l'arete courante */
		depart.addSuccesseur(curArr);
		}
	while (curArr != depart);
	}


	/* construit le graphe des aretes */

	private void construireGraphe()
	{
	/* pour chaque obstacle */
	/* 	pour chaque arete de cet obstacle dont il est l'origine */
	/* 		parcours  gauche &  droite */
	/* 		les aretes successeurs sont celles dont le x est != */

	for (int i = 0; i < obstacles.size(); i++)
		{
		Obstacle o = (Obstacle) obstacles.get(i);

		for (int j = 0; j < o.nbArete(); j++)
			{
			Arete a = o.getArete(j);

			/* pour les aretes dont l'objet o est l'origine */
			if (o == a.getObsSrc())
				{
				/* parcours devant */
				rechSuccesseurs(a, 1);

				/* parcours derrire si pas sur le cadre */
				/* (sinon on trouve 2 fois chaque successeurs) */
				if (!(o instanceof Cadre))
					rechSuccesseurs(a, -1);
				}
			}
		}
	}


	/* postcondition : l'origne du chemin est sur le cadre (idem pour la fin du chemin)
	 */
	/*@ private normal_behavior
	  @ ensures ( (((Arete) chemin.get(0)).getObsSrc() instanceof Cadre) &&
	  @			(((Arete) chemin.get(0)).getObsDest() instanceof Cadre) &&
	  @ 		(((Arete) chemin.get(chemin.size()-1)).getObsSrc() instanceof Cadre) &&
	  @			(((Arete) chemin.get(chemin.size()-1)).getObsDest() instanceof Cadre));
	  @
	  @ assignable \nothing;
	  @*/
	public void getChemin(Arete courante)
	{
	while (courante != null)
		{
		chemin.add(courante);
		courante = courante.getPlusLongueADroite();
		}
	}


	/* invariant : les points du chemin sont bien dans le cadre
	 */
	/*@ private invariant (\forall int i; 0 <= i && i < milieux.size();
	  @ 				((Point) milieux.get(i)).getX() >= 0 && ((Point) milieux.get(i)).getY() >= 0 &&
	  @					((Cadre) obstacles.get(0)).getLongueur() >= ((Point) milieux.get(i)).getX() &&
	  @					((Cadre) obstacles.get(0)).getHauteur() >= ((Point) milieux.get(i)).getY());
	  @*/
	private void getMilieux()
	{
	for (int i = 0; i < chemin.size(); i++)
		{
		milieux.add(((Arete) chemin.get(i)).getMilieu());
		System.out.println(((Point) milieux.get(i)).toString());
		}
	}


	private void transmetRCX()
	{
	try	{
		dos.writeInt( milieux.size() );
		int somme = 0;

		for (int i = 0; i < milieux.size(); ++i)
			{
			dos.writeInt( (int) ((Point) milieux.get(i)).getX() );
			somme += (int) ((Point) milieux.get(i)).getX();
			dos.writeInt( (int) ((Point) milieux.get(i)).getY() );
			somme += (int) ((Point) milieux.get(i)).getY();
			}
		System.out.println("somme calcule : " + somme + "  somme reue : " + dis.readInt());
		}
	catch (Exception e)
		{}
	}


	private void recupereRCX()
	{
/*	int num = 0;	
	
	for(int i = 0 ;i < milieux.size(); ++i)
		{
		try
			{
			num = dis.readInt()+1;
			}
		catch (Exception e)
			{}
		System.out.println("On a passe le " + num + "eme point");	
 		}
	System.out.println("On est arrive!!");
*/ 	}


	public static void main(String[] params)
	{
	AlgoRechercheChemin arc = new AlgoRechercheChemin();
	}
}

