
import java.util.*;
import java.awt.*;




class Arete
{
	/*
	une arete est constitue de :
	  . l'objet de dpart
	  . l'origine de l'arete
	  . l'objet de destination
	  . sa longueur
	  . son sens
	*/

	Obstacle dest;	// l'obstacle de destination de l'arete
	Obstacle src;	// l'obstacle d'origine de l'arete
	Point origine;	// le point d'origine de l'arete
	int longueur;	// sa longueur
	int sens;		// l'orientation de l'arete

	Vector successeurs;	// la liste des aretes successeurs


	Arete(ObstacleReference origine, ObstacleReference destination)
	{
	this.src = origine.getObstacle();
	this.dest = destination.getObstacle();
	this.origine = origine.getPoint();

	this.sens = (int) destination.getPoint().getY();
	this.successeurs = new Vector();

	int y = 0;
	if (dest instanceof Canette)
		{
		Canette c = (Canette) dest;

		longueur = pythagore((int) Math.abs(this.origine.getX() - c.getCentre().getX()), c.getRayon());
		y = (int) c.getCentre().getY();
		}
	else
		{
		Cadre c = (Cadre) dest;

		longueur = 0;
		if (pointeVersLeHaut())
			y = c.getHauteur();
		else
			y = 0;
		}
	this.longueur = (int) Math.abs(this.origine.getY() - y) - longueur;
	}


	private /*@ pure */ int pythagore(int a, int c)
	{
	return (int) Math.sqrt(c*c - a*a);
	}


	public /*@ pure */ int getLongueur()
	{
	return longueur;
	}


	public /*@ pure */ Point getMilieu()
	{
	return new Point((int) origine.getX(), (int) origine.getY() - sens * longueur/2);
	}


	public /*@ pure */ Point getOrigine()
	{
	return origine;
	}


	public /*@ pure */ Obstacle getObsSrc()
	{
	return src;
	}


	public /*@ pure */ Obstacle getObsDest()
	{
	return dest;
	}


	public /*@ pure */ boolean pointeVersLeHaut()
	{
	return (sens == -1);
	}


	public /*@ pure */ boolean pointeVersLeBas()
	{
	return (sens == 1);
	}


	public /*@ pure */ void addSuccesseur(Arete a)
	{
	if (a != this)
		{
		successeurs.add(a);
		}
	}


	public /*@ pure */ boolean bonneLongueur(Arete courante)
	{
	int distCentres;
	int largeurRobot = 25;
	int rayon1, rayon2;
	Point point1, point2;


	if (courante.getObsSrc() instanceof Cadre)
		{
		point1 = courante.getOrigine();
		rayon1 = 0;
		}
	else
		{
		point1 = courante.getObsSrc().getCentre();
		rayon1 = courante.getObsSrc().getRayon();
		}

	if (courante.getObsDest() instanceof Cadre)
		{
		if (courante.pointeVersLeHaut())
			point2 = new Point((int) point1.getX(), ((Cadre) courante.getObsDest()).getHauteur());
		else
			point2 = new Point((int) point1.getX(), (int) courante.getObsDest().getGauche().getY());
		rayon2 = 0;
		}
	else
		{
		point2 = courante.getObsDest().getCentre();
		rayon2 = courante.getObsDest().getRayon();
		}

	distCentres = (int) Math.sqrt(Math.pow(Math.abs(point1.getX()-point2.getX()),2)+ Math.pow(Math.abs(point1.getY()-point2.getY()),2));

	return (distCentres >= largeurRobot + rayon1 + rayon2);
 	}


	/* on choisit l'arete de droite la plus grande
	 */
	/* postcondition : il n'existe pas d'arete  droite, ou on retourne une arete de longueur suffisante pour pouvoir passer
	 */
	/*@ ensures (\result == null || bonneLongueur(\result));
	  @*/
	public /*@ pure */ Arete getPlusLongueADroite()
	{
	Arete max = null;
	Arete courant = null;

	for (int i = 0; i < successeurs.size(); i++)
		{
		courant = (Arete) successeurs.get(i);
		if (courant.getOrigine().getX() > origine.getX())
			if ((max == null) || (courant.getLongueur() > max.getLongueur()))
				max = courant;
		}

	return max;
	}


	public int compareTo(Object a, Obstacle o)
	{
	Arete arete1 = this;
	Arete arete2 = (Arete) a;
	boolean arete1VersLeHaut;
	boolean arete2VersLeHaut;


	if (arete1.getObsSrc() != o)
		arete1VersLeHaut = !arete1.pointeVersLeHaut();
	else
		arete1VersLeHaut = arete1.pointeVersLeHaut();

	if (arete2.getObsSrc() != o)
		arete2VersLeHaut = !arete2.pointeVersLeHaut();
	else
		arete2VersLeHaut = arete2.pointeVersLeHaut();

	/* les 2 aretes sont places sur le haut de l'obstacle */
	if (arete1VersLeHaut && arete2VersLeHaut)
		{
		if (arete1.getOrigine().getX() < arete2.getOrigine().getX())
			return 1;
		else if (arete1.getOrigine().getX() > arete2.getOrigine().getX())
			return -1;
		else
			return 0;
		}
	/* les 2 aretes sont places sur le bas de l'obstacle */
	else if (!arete1VersLeHaut && !arete2VersLeHaut)
		{
		if (arete1.getOrigine().getX() < arete2.getOrigine().getX())
			return -1;
		else if (arete1.getOrigine().getX() > arete2.getOrigine().getX())
			return 1;
		else
			return 0;
		}
	else
		{
		if (arete1VersLeHaut && !arete2VersLeHaut)
			return 1;
		else if (!arete1VersLeHaut && arete2VersLeHaut)
			return -1;
		}
	return 100;
	}


	public /*@ pure */ String toString()
	{
	if (pointeVersLeHaut())
		return ("src: " + src.toString() + " [" + origine.getX() + "," + origine.getY() + "] " + " dest: " + dest.toString() + " l: " + longueur + " s: haut");
	else
		return ("src: " + src.toString() + " [" + origine.getX() + "," + origine.getY() + "] " + " dest: " + dest.toString() + " l: " + longueur + " s: bas");
	}
}

