
import josx.platform.rcx.*;
import josx.robotics.*;
import josx.util.*;



class NewNavigator implements Navigator
{
	private static final int SPEED = 7;	/* vitesse initiale des 2 moteurs */

	private int speedLeft = SPEED;		/* vitesse du moteur gauche plus petite que droit car ajustement pour aller a peu pres droit */
	private int speedRight = SPEED;		/* vitesse du moteur droit */
	private int tL;				/* temps pour faire un clic droit */
	private int tR;				/* temps pour faire un clic gauche */


	private Motor leftMotor;		/* moteur gauche */
	private Sensor leftContactSensor;	/* capteur gauche */
	private ClicListener leftClicListener;	/* compteur de clic gauche */

	private Motor rightMotor;		/* moteur droit */
	private Sensor rightContactSensor;	/* capteur droit */
	private ClicListener rightClicListener;	/* compteur de clic droit */

	Timer timer;				/* un timer */
	private int millisParCentimetre;	/* le temps pour avancer de 1 cm */
	private int aC;				/* le temps pour avancer de 1 */


	private Point position;			/* position du robot par rapport  sa position initiale */
	private double orientation;		/* angle en degr */


	private static final int d = 3;		/* distance entre 2 clics */
	private static final int a = 27;	/* angle entre 2 clics */


	/* Construction du navigateur du robot*/
	public NewNavigator(Motor mG, Motor mD, Sensor sG, Sensor sD, int tClic, int aClic)
	{
	leftMotor = mG;
	rightMotor = mD;
	leftContactSensor = sG;
	rightContactSensor = sD;
	initNavigator();
	leftClicListener = new ClicListener();
	rightClicListener = new ClicListener();
	leftContactSensor.addSensorListener(leftClicListener);
	rightContactSensor.addSensorListener(rightClicListener);
	this.millisParCentimetre = tClic / d;
	this.aC = aClic / a;
	position = new Point();
	orientation = 0.0;
	}



	/*permet d'initialiser le mode des capteurs et la vitesse des moteurs*/
	private void initNavigator()
	{
	leftContactSensor.setTypeAndMode(1, 0x20);
	leftContactSensor.activate();
	rightContactSensor.setTypeAndMode(1, 0x20);
	rightContactSensor.activate();
	leftMotor.setPower(speedLeft);
	rightMotor.setPower(speedRight);
	}



	/* permet de calculer le temps total d'avancement de distance du robot durant nbClics*/
	int travCalcTimeReferenceValue(int nbClics, int distance)
	{
	int t, oldTime, newTime;

	leftClicListener.resetNbClic();
	rightClicListener.resetNbClic();
	oldTime = (int)System.currentTimeMillis();
	if (distance > 0) forward();
	else backward();
	while (leftClicListener.getNbClic() != nbClics) {}
	leftMotor.stop();
	rightMotor.stop();
	newTime = (int)System.currentTimeMillis();
	t = (newTime - oldTime);
	return t;
	}



	/* permet de calculer le temps total de rotation angle du robot pdt nbClic*/
	int rotCalcTimeReferenceValue(int nbClics, int angle)
	{
	int t, oldTime, newTime;

	leftClicListener.resetNbClic();
	rightClicListener.resetNbClic();
	oldTime = (int)System.currentTimeMillis();
	if (angle > 0) turnleft();
	else turnright();
	while (leftClicListener.getNbClic() != nbClics)
		{}
	stop();
	newTime = (int)System.currentTimeMillis();
	t = (newTime - oldTime);

	return t;
	}



	/* fait reculer le robot*/
	public void backward()
	{
	leftMotor.backward();
	rightMotor.backward();
	}



	/*fait avancer le robot*/
	public void forward()
	{
	leftMotor.forward();
	rightMotor.forward();
	}



	/* fait tourner le robot sur sa gauche */
	public void turnleft()
	{
	leftMotor.backward();
	rightMotor.forward();
	}



	/* fait tourner le robot sur sa droite */
	public void turnright()
	{
	leftMotor.forward();
	rightMotor.backward();
	}



	/*Le robot s'arrete*/
	public void stop()
	{
	leftMotor.stop();
	rightMotor.stop();
	}



	/*Le robot tourne d'un angle an */
	public void rotate(float an)
	{
	int nbC;		/*nbC est le nbre de clic a faire pour avancer d'une distance*/
	int angleReste;	/*la distance restant a avancer*/
	int t1L, t1R;	/*le temps que le robot a mis pour le premier clic s'il n'etait pas synchronise */
	boolean isActivate;
	int oldTime, newTime;

	int angle = (int) an;

	if (angle == 0) return;
	nbC = (int) Math.floor(Math.abs(angle)/a);
	angleReste = Math.abs(angle) % a;
	if (angle > 0)
		turnleft();
	else
		turnright();

	if (nbC >= 2)
		{
		oldTime = (int)System.currentTimeMillis();
		isActivate = leftContactSensor.readBooleanValue();
		while(isActivate == false)
			{
			isActivate = leftContactSensor.readBooleanValue();
			}
		newTime = (int)System.currentTimeMillis();
		t1L = (newTime - oldTime);

		rotPause(angle);
		aC = rotCalcTimeReferenceValue(nbC-1, angle)/((nbC-1)*a);

		rotPause(angle);
		try	{Thread.currentThread().sleep((a + angleReste)* aC - t1L);}	catch (Exception e){}
	 	stop();
		}
	else
		{
		try	{Thread.currentThread().sleep(Math.abs(angle * aC));} catch (Exception e){}
	 	stop();
		}
	orientation += an;
	while (orientation >= 180.0)
		orientation -= 360.0;
	while (orientation <= -180.0)
		orientation += 360.0;
	}



	/*Le robot avance de distance*/
	public void travel(int distance)
	{
	int nbC;/*nbC est le nbre de clic a faire pour avancer d'une distance*/
	int distReste;/*la distance restant a avancer*/
	int t1L, t1R;/*le temps que le robot a mis pour le premier clic s'il n'etait pas synchronise */
	boolean isActivate;
	int oldTime, newTime;

	if (distance == 0) return;
	nbC = Math.abs(distance)/d;
	distReste = distance % d;
	if (distance > 0)
		forward();
	else
		backward();

	if (nbC >= 2)
		{
		/* avance de t avant la premire touche (< 1 clic) */
		oldTime = (int)System.currentTimeMillis();
		leftClicListener.resetNbClic();
		while (leftClicListener.getNbClic() != 1) {}
		newTime = (int)System.currentTimeMillis();
		t1L = (newTime - oldTime);
		travPause(distance);

		/* avance de nbClic - 1 */
		millisParCentimetre = travCalcTimeReferenceValue(nbC-1, distance)/((nbC-1)*d);
		travPause(distance);

		/* termine le clic commenc au dbut + termine le mouvement */
		try
			{
			Thread.currentThread().sleep((d + distReste)* millisParCentimetre - t1L);
	 		}
	 	catch (Exception e) {}
	 	stop();
		}
	else
		{
		try
			{
			Thread.currentThread().sleep(Math.abs(distance * millisParCentimetre));
	 		}
	 	catch (Exception e){}
	 	stop();
		}
	position.setX(distance * Math.cos((float) orientation) + position.getX());
	position.setY(distance * Math.sin((float) orientation) + position.getY());
	}



	/*pause pour travel*/
	private void travPause(int distance)
	{
	stop();
	try
		{
		Thread.currentThread().sleep(0);
		}
	catch (Exception e) {}
	if (distance > 0)
		forward();
	else
		backward();
	}



	/*pause pour rotation*/
	private void rotPause(int angle)
	{
	stop();
	try
		{
		Thread.currentThread().sleep(0);
		}
	catch (Exception e) {}
	if (angle > 0)
		turnleft();
	else
		turnright();
	}



	/*Fait tourner le robot jusqu'a l'angle angle*/
	public void gotoAngle(float angle)
	{
	rotate(angle - (float)orientation);
	}



	/*Le robot se deplace jusqu'au point (x,y)*/
	public void gotoPoint(float x, float y)
	{
	int norme;
	double angle_rot;
	Point v = new Point();

	norme = (int) Math.sqrt((x - position.getX())*(x - position.getX()) + (y -position.getY())*(y - position.getY()));
	v.setX((x - position.getX())/norme);
	v.setY((y - position.getY())/norme);
	angle_rot = Math.toDegrees(Math.acos(v.getX()));
	if (v.getX() == -1)
		angle_rot=180;
	else if (v.getX() < 0)
		{
	        angle_rot = Math.toDegrees(Math.acos((float) - v.getX()));
			if (v.getY() < 0)
				angle_rot = (angle_rot - 180);
			else
				angle_rot = (180 - angle_rot);
		}
	else
		{
	     	angle_rot = Math.toDegrees(Math.acos(v.getX()));
	     	if (v.getY() < 0) angle_rot = -angle_rot;
		}

	try{Thread.sleep(1000);} catch (Exception e) {}
	gotoAngle((float) angle_rot);

	try{Thread.sleep(1000);} catch (Exception e) {}
	travel(norme);

	try{Thread.sleep(1000);} catch (Exception e) {}
	position.setX(x);
	position.setY(y);
	}



	/* retourne l'angle actuel du robot*/
	public float getAngle()
	{
	return (float) (orientation % 360);
	}


	/* retourne la position X du robot*/
	public float getX()
	{
	return (float) position.getX();
	}



	/* retourne la position Y du robot*/
	public float getY()
	{
	return (float) position.getY();
	}
}




/* permet de compter le nombre de clics d'un des capteurs */
/* et gre le temps moyen entre 2 clics  */

class ClicListener implements SensorListener
	{
	int nbClic;		/* compteur du nombre de clics */
	int nbClicTotal;	/* le nombre de clics total */
	int tempsTotal;		/* temps total */
	int lastTime;

	int oldTime;
	int newTime;



	public ClicListener()
	{
	super();
	nbClic = 0;
	nbClicTotal = 0;
	tempsTotal = 0;		/* A INITIALISER AVEC UNE VALEUR PASSE EN PARAM */
	lastTime = 0;
 	}



	/* clic */
	public void stateChanged(Sensor capt, int oldVal, int newVal)
	{
	if (newVal == 1 && oldVal == 0)	/* dbut d'un clic */
		{
		if (nbClic == 0)			/* dans le cas ou on commence un mouvement, on ne recalcule pas le temps moyen */
			{
			oldTime = (int) System.currentTimeMillis();
			}
		else
			{
			newTime = (int) System.currentTimeMillis();
			lastTime = newTime - oldTime;
			tempsTotal += lastTime;
			oldTime = newTime;
			}
		nbClic++;
		nbClicTotal++;
		}
 	}



	/* remet le compteur de clic  0 */
	public void resetNbClic()
	{
		nbClic = 0;
 	}



 	/* retourne le nbClic */
 	public int getNbClic()
	{
	return nbClic;
	}



 	/* retourne le temps moyen entre 2 clics */
 	public int getTempsMoyen()
	{
	return tempsTotal / nbClicTotal;
	}



	/*Retourne le temps entre 2 clics*/
	public int getLastTime()
	{
	return lastTime;
 	}
}
